use actix_web::error::ErrorNotFound;
use actix_web::{error::ErrorInternalServerError, web, HttpRequest, HttpResponse, Result};
use log::{error, warn};

use crate::app_state::AppState;
use crate::handlers::serve_file::serve_file;
use crate::s3_utils::{check_file_exists, load_file_from_s3, upload_to_s3};
use crate::thumbnail::{find_closest_width, parse_file_path, thumbdata_save};

/// Обработчик для скачивания файла и генерации миниатюры, если она недоступна.
pub async fn proxy_handler(
    req: HttpRequest,
    requested_res: web::Path<String>,
    state: web::Data<AppState>,
) -> Result<HttpResponse, actix_web::Error> {
    warn!("\t>>>\tGET {}", requested_res);
    let normalized_path = if requested_res.ends_with("/webp") {
        requested_res.replace("/webp", "")
    } else {
        requested_res.to_string()
    };

    // парсим GET запрос
    let (base_filename, requested_width, extension) = parse_file_path(&normalized_path);
    warn!("detected file extension: {}", extension);
    warn!("base_filename: {}", base_filename);
    warn!("requested width: {}", requested_width);
    let ext = extension.as_str().to_lowercase();
    warn!("normalized to lowercase: {}", ext);
    let filekey = format!("{}.{}", base_filename, &ext);
    warn!("filekey: {}", filekey);
    let content_type = match ext.as_str() {
        "jpg" | "jpeg" => "image/jpeg",
        "png" => "image/png",
        "webp" => "image/webp",
        "gif" => "image/gif",
        "mp3" => "audio/mpeg",
        "wav" => "audio/x-wav",
        "ogg" => "audio/ogg",
        "aac" => "audio/aac",
        "m4a" => "audio/m4a",
        "flac" => "audio/flac",
        _ => {
            error!("unsupported file format");
            return Err(ErrorInternalServerError("unsupported file format"));
        },
    };

    warn!("content_type: {}", content_type);

    let shout_id = match req.query_string().contains("s=") {
        true => req.query_string().split("s=").collect::<Vec<&str>>().pop().unwrap_or(""),
        false => ""
    };

    return match state.get_path(&filekey).await {
        Ok(Some(stored_path)) => {
            warn!("stored_path: {}", stored_path);
            // Проверяем, существует ли файл в Storj
            if check_file_exists(&state.storj_client, &state.bucket, &stored_path).await? {
                if content_type.starts_with("image") {
                    if requested_width == 0 {
                        return serve_file(&stored_path, &state, shout_id).await;
                    } else {
                        // Находим ближайшую ширину для миниатюры
                        let closest: u32 = find_closest_width(requested_width as u32);
                        let thumb_filename = &format!("{}_{}.{}", base_filename, closest, ext);

                        // Проверяем, существует ли уже миниатюра в Storj
                        match check_file_exists(&state.storj_client, &state.bucket, thumb_filename).await {
                            Ok(true) => {
                                warn!("serve existed thumb file: {}", thumb_filename);
                                return serve_file(thumb_filename, &state, shout_id).await;
                            },
                            Ok(false) => {
                                // Миниатюра не существует, создаем и загружаем её
                                if let Ok(filedata) = load_file_from_s3(&state.storj_client, &state.bucket, &stored_path).await {
                                    warn!("generate new thumb files: {}", stored_path);
                                    warn!("{} bytes", filedata.len());

                                    // Генерируем миниатюру и сохраняем
                                    if let Ok(_) = thumbdata_save(filedata.clone(), &state, &filekey, content_type.to_string()).await {
                                        warn!("serve new thumb file: {}", thumb_filename);
                                        return serve_file(thumb_filename, &state, shout_id).await;
                                    } else {
                                        error!("cannot generate thumbnail");
                                        return Err(ErrorInternalServerError("cannot generate thumbnail"));
                                    }
                                } else {
                                    error!("cannot load file from Storj to generate thumbnail");
                                    return Err(ErrorInternalServerError("cannot generate thumbnail"));
                                }
                            }
                            Err(e) => {
                                error!("ошибка при проверке существования миниатюры: {}", e);
                                return Err(ErrorInternalServerError("failed to load thumbnail"));
                            }
                        }
                    }
                }
                // Если файл не изображение, продолжаем обработку
                warn!("file is not an image");
            }

            // we need to download what stored_path keeping in aws
            return match load_file_from_s3(&state.aws_client, &state.bucket, &stored_path).await
            {
                Ok(filedata) => {
                    warn!("download stored_path from aws: {:?}", stored_path);
                    if let Err(e) = upload_to_s3(
                        &state.storj_client,
                        &state.bucket,
                        &filekey,
                        filedata.clone(),
                        content_type,
                    )
                    .await {
                        error!("cannot upload to storj: {}", e);
                    } else {
                        warn!("file {} uploaded to storj", filekey);
                    }
                    let _ = thumbdata_save(filedata.clone(), &state, &filekey, content_type.to_string())
                        .await;

                    Ok(HttpResponse::Ok().content_type(content_type).body(filedata))
                }
                Err(err) => {
                    error!("cannot download {} from aws: {}", stored_path, err);
                    Err(ErrorInternalServerError(err))
                },
            };
        }
        Ok(None) => {
            warn!("cannot find stored path for: {}", filekey);
            let ct_parts = content_type.split("/").collect::<Vec<&str>>();
            let filepath = format!("production/{}/{}.{}", ct_parts[0], base_filename, extension); // NOTE: original ext

            let exists_in_storj = check_file_exists(&state.storj_client, &state.bucket, &filepath).await?;

            if exists_in_storj {
                warn!("file {} exists in storj, try to generate thumbnails", filepath);

                match load_file_from_s3(&state.aws_client, &state.bucket, &filepath).await {
                    Ok(filedata) => {
                        let _ = thumbdata_save(filedata.clone(), &state, &filekey, content_type.to_string()).await;
                    }
                    Err(e) => {
                        error!("cannot download {} from storj: {}", filekey, e);
                        return Err(ErrorInternalServerError(e));
                    }
                }
            } else {
                warn!("file {} does not exist in storj", filepath);
            }

            let exists_in_aws = check_file_exists(&state.aws_client, &state.bucket, &filepath).await?;
            if exists_in_aws {
                match load_file_from_s3(&state.aws_client, &state.bucket, &filepath).await {
                    Ok(filedata) => {
                        let _ = thumbdata_save(filedata.clone(), &state, &filekey, content_type.to_string())
                            .await;
                        if let Err(e) = upload_to_s3(
                            &state.storj_client,
                            &state.bucket,
                            &filekey,
                            filedata.clone(),
                            content_type,
                        )
                        .await {
                            warn!("cannot upload to storj: {}", e);
                        } else {
                            warn!("file {} uploaded to storj", filekey);
                        }
                        Ok(HttpResponse::Ok().content_type(content_type).body(filedata))
                    },
                    Err(e) => {
                        error!("cannot download {} from aws: {}", filepath, e);
                        Err(ErrorInternalServerError(e))
                    },
                }
            } else {
                error!("file {} does not exist in aws", filepath);
                Err(ErrorNotFound("file does not exist"))
            }
        },
        Err(e) => {
            error!("cannot get path from aws: {}", e);
            Err(ErrorInternalServerError(e))
        }
    }
}
